% Solve 3D convection-difussion equation on irregular geometry with partition 
% of unity collocation method where RBF interpolants are used as local interpolants.
%
% u_t = Lap*u + v . Grad u
%
% v = -c <1,1,1>
%
% Exact solution u(x,y,z,t) = A*exp(b*t - c*(x+y+z));
% A = any constant, b = any positive constant, c = sqrt(b/6);
%
% Use Diriclet boundary conditions taken from exact solution.
%
% Domain bounded by x^2+y^2+z^2 - sin(2*x)^2*sin(2*y)^2*sin(2*z)^2 = 1;

% This MATLAB code is released to support reproducible research for numerical 
% results in the manuscript:
% A radial basis function partition of unity collocation method for
% convection-diffusion equations arising in financial applications.
% Status: Submitted
% Ali Safdari-Vaighani, Alfa Heryudono, Elisabeth Larsson.

% Please report bugs, questions, and suggestions to
% Alfa Heryudono
% Department of Mathematics
% UMass Dartmouth
% e-mail: aheryudono@umassd.edu
% website: http://www.math.umassd.edu/~aheryudono/

% When submitting bugs reports, please include your name and affiliation
% so that we can acknowledge you properly in our papers. If you find this 
% code useful for your research, do not hesitate to acknowledge us too :)

% What you need to run this code:
% MATLAB: tested on 2013a and above with parallel computing toolbox.
% Kd-tree package from Andrea Tagliasacchi from MATLAB files exchange.
% You need to turn on matlabpool.

% Parallel version.
clear all; close all;
warning('off','all')

if matlabpool('size') == 0
    disp('You ask for Parallel Computation but matlabpool is not on. Type matlabpool')
    keyboard;
end

% ============ Input Parameter ============

% Convection diffusion parameter
b = 0.1;
c = sqrt(b/6);

% Turn this on if you need to see some plots
flagPLOT = true;

% Set dr to be greater than 1 if you want to remove disks whose centers
% are outside dr*r(theta) distance from the boundary.
dr = 1;

% Scale radius overlapping factor 
rof = 1;

% Levelinit
levelinit = 4; % Set boxes up to level = levelinit

% Set bounding box covering the entire domain. 
% Its size is determined based on partition of radius rb
%rb = 0.175
%rb = 0.125 %% needed
r3 = ((3/4)^(3/2))*(8/8^levelinit); %needed
rb = r3^(1/3);


% Load RBF points
% Note that .mat files must contain
% 1. xc: column vectors of RBF nodes.
% 2. flagbdr: containing flag of boundary points.
% 3. eg: global shape parameter.

load RBFPts3DN2046optim.mat


% Number of points inside each partition
N = size(xc,1);
nn = round(r3*10*N);
nn = 3*nn;

% ============ End of Input Parameter ============

% Cover domain with balls of radius rb.
% The largest bounding box covering the entire domain will be determined
% from rb
x1 = -(2^levelinit)*rb; y1 = x1; z1 = x1;
x2 = -x1; y2 = x2; z2 = x2; 
if x2 < 0.9
    disp('Bounding box too small')
    disp(x2)
    keyboard
end

% Create overlapping disks to cover [x1,x2]x[y1,y2]. The square domain is 
% subdivided into boxes using quadtree decompositions. The midpoints of 
% those boxes are used as centers of the overlapping disks. The partitions
% are organized using quadtree data structure.

[ibox] = initboxes3d(x1,x2,y1,y2,z1,z2,rof); % Initial boxes.
% Refine boxes up to level = levelinit
icenter = deal(true);
for i=1:levelinit
    [ibox,inewres] = dorefine3d(ibox,icenter,rof);
    icenter = false(size(ibox));  icenter(inewres) = true;
end

%Turn off disks whose centers are outside the domain
w = cat(1,ibox.midpt);
partoff = (w(:,1).^2+w(:,2).^2+w(:,3).^2) > dr*(1 + ...
          sin(2*w(:,1)).^2.*sin(2*w(:,2)).^2.*sin(2*w(:,3)).^2);
icenter(partoff) = false;

switch flagPLOT
   case true
       [figh1,figh2] = boxplot3d(xc,flagbdr,ibox,icenter);
end

N = length(xc(:,1));
xi = num2cell(xc,2);
[data(1:N).ctr] = deal(xi{:}); clear xi;
[data.circind] = deal([]);
bdr = num2cell(flagbdr);
[data.isbdr]=deal(bdr{:});

% For each center, find which partitions it lies into. For each disk,
% find which centers that lie inside.
tree = kdtree_build( xc );
for i=find(icenter)
    q = ibox(i).midpt;
    idxs = kdtree_k_nearest_neighbors(tree, q, nn+1);
    r12 = sqrt((q(1)-xc(idxs(1:2),1)).^2 + (q(2)-xc(idxs(1:2),2)).^2 + ...
          (q(3)-xc(idxs(1:2),3)).^2);
    ibox(i).R = 0.5*sum(r12);
    idxs(1) = []; ibox(i).indlocpts = idxs;
    
    for j=1:length(idxs)
        data(idxs(j)).circind = [data(idxs(j)).circind i];
    end
end

% Check if there are uncovered points
% Find indices of closest partitions and stagger those partitions
% to get cover those uncovered points
uncovered = arrayfun(@(x)(isempty(x.circind)),data);
while any(uncovered)
    disp('Cover uncovered points by staggering closest partitions.')
    
    flagstag = false(size(ibox));
    flagstagR = spalloc(1,length(ibox),length(uncovered));
    flagstagidx = spalloc(1,length(ibox),length(uncovered));
    for i=find(uncovered)
        [idxref,Rmin] = getclosestsphereind(data(i).ctr,ibox,length(ibox),icenter);
        flagstag(idxref) = true;
        if flagstagR(idxref) < Rmin
            flagstagR(idxref) = Rmin;
            flagstagidx(idxref) = i;
        end
    end
    
    % Cleaning
    for i=find(flagstag)
        idxs = ibox(i).indlocpts;
        ibox(i).indlocpts = [];
        
        for j=1:length(idxs)
            [tf,loc]=ismember(i,data(idxs(j)).circind);
            if tf
                data(idxs(j)).circind(loc) = [];
            end
        end
        
        % Shift centers
        xd = data(flagstagidx(i)).ctr;
        dv = xd - ibox(i).midpt;
        %dR = flagstagR(i) - ibox(i).R;
        %ibox(i).midpt = ibox(i).midpt + dR*dv/hypot(dv(1),dv(2));
        
        [adv,idx] = max(abs(dv));
        flagst = false(1,3); flagst(idx)=true;
        dD = adv-0.5*ibox(i).dimen(flagst);
        ibox(i).midpt(flagst)= ibox(i).midpt(flagst) + 0.5*sign(dv(flagst))*dD;
        ibox(i).midpt(~flagst) = ibox(i).midpt(~flagst) + 0.5*sign(dv(~flagst))*dD;
        ibox(i).dimen(flagst) =  ibox(i).dimen(flagst) + dD;
        ibox(i).dimen(~flagst) = ibox(i).dimen(flagst);
        ibox(i).R = sqrt(sum((0.5*ibox(i).dimen).^2));
    end
    
    % Update
    for i=find(flagstag)
        q = ibox(i).midpt;
        idxs = kdtree_ball_query(tree, q, ibox(i).R);
        ibox(i).indlocpts = idxs;
        
        for j=1:length(idxs)
            data(idxs(j)).circind = [data(idxs(j)).circind i];
        end
    end

    uncovered = arrayfun(@(x)(isempty(x.circind)),data);
end
% Delete kdtree
kdtree_delete( tree );

switch flagPLOT
    case true
        [figh1,figh2] = boxplot3d(xc,flagbdr,ibox,icenter);
        
%         for i=1:8
%             view(i*45,0);
%             pause
%         end
%         view(3)
end

% Check if there are still uncovered points (if any)
uc = arrayfun(@(x)(numel(x.circind)),data);
if nnz(uc) < N
    disp('There are uncovered points')
    flagbad = uc == 0;
    plot3(xc(flagbad,1),xc(flagbad,2),xc(flagbad,3),'o','markerfacecolor','k')
    keyboard
end

nloc = arrayfun(@(x)(numel(x.indlocpts)),ibox);
%nlocmin = min(nonzeros(nloc));
nlocmax = max(nonzeros(nloc));

switch flagPLOT
    case true
        figure;hist(nonzeros(nloc),min(nonzeros(nloc)):max(nonzeros(nloc)));
        drawnow
end

spmd
    % Preallocate matrices to every worker
    % Pre-allocate sparse matrix
    L = spalloc(N,N,nlocmax*N);
    
    locindices = 1:N;
    %locindices = codistributed(locindices,'convert'); %2008b 2009a
    %locindices = localPart(locindices); % 2008b 2009a
    locindices = codistributed(locindices); %2009b - 2014a
    locindices = getLocalPart(locindices);  %2009b - 2014a 
end

disp('Assemble differentiation matrices in parallel')

% Get differentiation matrices
Np = sum(icenter);
tic
% Fill in weights in sparse matrices (column-wise is faster than row-wise).
% Don't forget to transpose.
spmd
    warning('off','all')
    for k=1:length(locindices)
        i = locindices(k);
        activecirc = data(i).circind;
        [puw,puwx,puwy,puwz,puwxx,puwyy,puwzz] = puweight(xc(i,:),activecirc,ibox);
        for j=1:length(activecirc)
            activectr = xc(ibox(activecirc(j)).indlocpts,:);
            
            % Scale activectr and xc(i,:) to unit circle.
            sactivectr = zeros(size(activectr));
            sactivectr(:,1) =(activectr(:,1)-ibox(activecirc(j)).midpt(1))/ibox(activecirc(j)).R;
            sactivectr(:,2) =(activectr(:,2)-ibox(activecirc(j)).midpt(2))/ibox(activecirc(j)).R;
            sactivectr(:,3) =(activectr(:,3)-ibox(activecirc(j)).midpt(3))/ibox(activecirc(j)).R;
            sxc = zeros(size(xc(i,:)));
            sxc(1) = (xc(i,1)-ibox(activecirc(j)).midpt(1))/ibox(activecirc(j)).R;
            sxc(2) = (xc(i,2)-ibox(activecirc(j)).midpt(2))/ibox(activecirc(j)).R;
            sxc(3) = (xc(i,3)-ibox(activecirc(j)).midpt(3))/ibox(activecirc(j)).R;
            
            %Find weights using regular RBF (In this case use GA-RBF)
            epscaled = ibox(activecirc(j)).R*eg;
            [weightu,weightux,weightuy,weightuz,weightLap] = gaweights(epscaled,sactivectr,sxc);
            weightLap = weightLap*(1/ibox(activecirc(j)).R)^2;
            weightux = weightux/ibox(activecirc(j)).R;
            weightuy = weightuy/ibox(activecirc(j)).R;
            weightuz = weightuz/ibox(activecirc(j)).R;
            %---------------------------
            
            activeind = ibox(activecirc(j)).indlocpts;
            
            L(activeind,i) = L(activeind,i) + puw(j)*weightLap(:) + ...
                2*(puwx(j)*weightux(:) + puwy(j)*weightuy(:) + puwz(j)*weightuz(:)) + ...
                (puwxx(j) + puwyy(j) + puwzz(j))*weightu(:);
            
            L(activeind,i) = L(activeind,i) + (-c)*...
                (puwx(j)*weightu(:)+puw(j)*weightux(:));
            
            L(activeind,i) = L(activeind,i) + (-c)*...
                (puwy(j)*weightu(:)+puw(j)*weightuy(:));
            
            L(activeind,i) = L(activeind,i) + (-c)*...
                (puwz(j)*weightu(:)+puw(j)*weightuz(:));
            
        end
        clc;fprintf('Compute weights for center %d of %d \n',i,N);
    end
end

spmd
    L = gplus(L);
end
L = L{1}; L = L.'; % Transplose L

L = L(~flagbdr,:);
Ladd = L(:,flagbdr);
L = L(:,~flagbdr);
toc

disp('Do time-stepping process')

% Exact solution.
uexact = @(t,x) exp(b*t-c*(x(:,1)+x(:,2)+x(:,3)));

% Let's march in time.
convdiff3D = @(t,u,L,Ladd,xc,flagbdr,b,c) L*u + Ladd*exp(b*t-c*(xc(flagbdr,1)+xc(flagbdr,2)+xc(flagbdr,3)));
Jconvdiff3D = @(t,u,L,Ladd,xc,flagbdr,b,c) L;
u0 = uexact(0,xc);
u0 = u0(~flagbdr);
options = odeset('RelTol',1e-6,'AbsTol',1e-5,'Jacobian',@(t,u)Jconvdiff3D(t,u,L,Ladd,xc,flagbdr,b,c));
tic
[T,U]=ode15s(@(t,u)convdiff3D(t,u,L,Ladd,xc,flagbdr,b,c),[0:0.5:10],u0,options);
toc

U = U.';
maxerr = zeros(size(T));
for i=1:length(T)
    maxerr(i) = max(abs(U(:,i)-uexact(T(i),xc(~flagbdr,:))));
end

figure;
semilogy(T,maxerr,'-o','markerfacecolor','k')
xlim([-0.1 11]); ylim([1e-6 1e-1])
